/***************************************************************************
 *   fadprogressd -- daemon to send progress of FAD(TM) calculations       *
 *   Copyright (C) 2005 by Eric Smith                                      *
 *   Contact information:  MerePeer at www.free-dc.org                     *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   The GPL can usually be found at http://www.gnu.org/licenses/gpl.txt   *
 ***************************************************************************/

// linux compile line:  gcc -O2 -o fadprogressd fadprogressd.c
// win32 compile line:  gcc -O2 -o fadprogressd fadprogressd.c -lws2_32

// version 1.00 -  4/06/05
//    - initial release
// version 0.04 -  3/16/05
//    - retreive and send CPU rating
//    - reworked for multi node server.dat
// version 0.03 -  3/9/05
//    - attempts at performance improvement
// version 0.02 -  2/27/05
//    - add win32 service install and remove code
// version 0.01 -  1/27/05 - creation

// includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>

#ifdef WIN32
	#include <windows.h>
	#include <winsock2.h>
#else
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <netdb.h>
#endif

// global defines
#define TRUE 1
#define FALSE 0
#define sLOGFILENAME "fadprogressd.log"
#define sHELPMSG "Usage:  fadprogressd\n \
	-sendto fadprogresshostname\n \
	[-dir dirname]\n \
	[-name thisnodename]\n \
	[-port hostportnum]\n \
	[-timer delayinsecs]\n \
	[-debug]\n \
	[-log]\n \
	[-logdir dirname]\n \
	OR, only on Windows(TM) these options are used without others:\n \
	-create_service\n \
	-delete_service\n"
#define sTIMEFORMAT "%Y/%m/%d %H:%M:%S"
#define sPROGRAM_HEADER "fadprogressd v1.00"
#define sPACKET_VERSION "D04"
#define sPACKET_DELIM "\t"
#define cPACKET_DELIM '\t'
#define sPACKET_TYPE_FAD "F"
#define sPACKET_TYPE_PROBLEM "A"
#define sSERVICE_NAME "dc_fad_pd"
#define sSERVICE_DISPLAY "DC FADProgress Daemon"
#define DAT_LINELEN 101
#define LINE_LEN 255


// cfg vars that can be overridden by args
char cfgFADDir[FILENAME_MAX] = {"./"};
char cfgLogDir[FILENAME_MAX] = {"./"};
char cfgServerName[256] = {0};
char cfgThisName[256] = {0};
int cfgDelay = { 10 };
unsigned short cfgServerPort = { 9031 };
int bDebug = { FALSE };
int bUseLogfile = { FALSE };

// globals
int bAsWinService = { FALSE };
int iCPURatingCheck = {0};
struct sockaddr_in udpServer;


#ifdef WIN32
	// service vars
	SERVICE_STATUS serviceStatus;
	SERVICE_STATUS_HANDLE hServiceStatus;

	void WINAPI serviceControlHandler(DWORD iOpcode);
	void stopService();
	void createService();
	void deleteService();

	#define SLEEP(n) sleep(n * 1000)
	SOCKET sd;

#else

	#define SLEEP(n) sleep(n)
	int sd;
#endif

// Prototypes
void serviceMain(int argc, char *argv[]);
static int doProgress();
static char * trim(char *);
static int report(char *);
static void appendSeparator(char *);
static void readIni(char *);
static void getCPURating(char *, char *);

/**********************************************************************************/
int main(int argc, char *argv[])
{
#ifdef WIN32
	SERVICE_TABLE_ENTRY serviceTable[] =
	{
		{ sSERVICE_NAME, (LPSERVICE_MAIN_FUNCTION) serviceMain },
		{ NULL, NULL }
	};

	if (argc > 1) // command line windows, not service mode
	{
		if (0 == strcmp(argv[1], "-create_service"))
		{
			createService();
		} else if (0 == strcmp(argv[1], "-delete_service"))
		{
			deleteService();
		} else {
			serviceMain(argc, argv);
		}
	}
	else // service mode
	{
		bAsWinService = TRUE;
		bUseLogfile = TRUE; // force the log file

		// create the ini filename by replacing the exe name's file extension
		char sIniFile[FILENAME_MAX];
		strcpy(sIniFile, argv[0]); // copy exe name
		char * spDot = strrchr(sIniFile, '.');
		if (spDot) *spDot = 0;
		strcat(sIniFile, ".ini");
		readIni(sIniFile); // read the ini file to get command values

		// tell the dispatcher to try starting our service
		if (!StartServiceCtrlDispatcher(serviceTable))
		{
			char sReport[128];
			sprintf(sReport, "ERROR from StartServiceCtrlDispatcher: %d", GetLastError());
			report(sReport);
			// reminder:  this error is reported IN THE LOG FILE, not to the console
		}
	}

#else // UNIX
	serviceMain(argc, argv);
#endif
}

/**********************************************************************************/
// workhorse fn of this application, called by main
void serviceMain(int argc, char *argv[])
{
#ifdef WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int iReturn;
#endif

	struct hostent *he;
	char sReport[256];
	char sNodeFile[FILENAME_MAX];
	int i;
	int iPrefixLen;
	long lThisNameSize = sizeof(cfgThisName);


#ifdef WIN32
	if (bAsWinService)
	{
		serviceStatus.dwServiceType = SERVICE_WIN32;
		serviceStatus.dwCurrentState = SERVICE_START_PENDING;
		serviceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
		serviceStatus.dwWin32ExitCode = 0;
		serviceStatus.dwServiceSpecificExitCode = 0;
		serviceStatus.dwCheckPoint = 0;
		serviceStatus.dwWaitHint = 0;

		hServiceStatus = RegisterServiceCtrlHandler(sSERVICE_NAME, serviceControlHandler);
		if (hServiceStatus == (SERVICE_STATUS_HANDLE)0)
		{
			sprintf(sReport,"ERROR calling RegisterServiceCtrlHandler: %d", GetLastError());
			report(sReport);
			return;
		}
	}

	// regardless of whether a service or not, setup winsock
	wVersionRequested = MAKEWORD(2, 2);

	iReturn = WSAStartup(wVersionRequested, &wsaData);
	if (iReturn != 0)
	{
		report("ERROR:  winsock dll v2.2 not found.");
  #ifdef WIN32
		if (bAsWinService)
			stopService();
		else
			exit(EXIT_FAILURE);
  #else
		exit(EXIT_FAILURE);
  #endif
	}

	// lower priority to barely-above-idle on windows
	iReturn = SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);
	if (iReturn == 0)
	{
		sprintf(sReport, "ERROR:  could not lower priority %d %d.", iReturn, GetLastError());
		report(sReport);
  #ifdef WIN32
		if (bAsWinService)
			stopService();
		else
			exit(EXIT_FAILURE);
  #else
		exit(EXIT_FAILURE);
  #endif
	}
#endif // WIN32

	// if cfgThisName was not already set by the readIni call (win32 svc only),
	// then populate a default val here using hostname.  It may still be overridden
	// with an argv value.
	if (strlen(cfgThisName) == 0) gethostname(cfgThisName, lThisNameSize);

	// handleany command line args
	for (i = 0; i < argc; i++)
	{
		if (0 == strcmp(argv[i], "-sendto")) {
			strcpy(cfgServerName, argv[++i]);
		} else if (0 == strcmp(argv[i], "-dir")) {
			strcpy(cfgFADDir, argv[++i]);
			appendSeparator(cfgFADDir);
		} else if (0 == strcmp(argv[i], "-logdir")) {
			strcpy(cfgLogDir, argv[++i]);
			appendSeparator(cfgLogDir);
		} else if (0 == strcmp(argv[i], "-name")) {
			strcpy(cfgThisName, argv[++i]);
		} else if (0 == strcmp(argv[i], "-port")) {
			cfgServerPort = atoi(argv[++i]);
		} else if (0 == strcmp(argv[i], "-timer")) {
			cfgDelay = atoi(argv[++i]);
		} else if (0 == strcmp(argv[i], "-debug")) {
			bDebug = TRUE;
		} else if (0 == strcmp(argv[i], "-log")) {
			bUseLogfile = TRUE;
		}
	}

	// server name is a required arg
	if (cfgServerName[0] == 0)
	{
		report(sHELPMSG);
		exit(EXIT_FAILURE);
	}

	// setup socket descriptor
	sd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sd == -1)
	{
		report("ERROR:  socket creation failed.");
		exit(EXIT_FAILURE);
	}

	// convert domain name into inet addr; does nothing if IP addr
	he = gethostbyname(cfgServerName);
	if (he == NULL)
	{
		sprintf(sReport, "ERROR:  unknown host '%s'", cfgServerName);
		report(sReport);
		exit(EXIT_FAILURE);
	}

	// setup server info
	udpServer.sin_family = AF_INET;
	udpServer.sin_port = htons(cfgServerPort);
	memcpy((char *) &udpServer.sin_addr.s_addr, he->h_addr_list[0], he->h_length);

#ifdef WIN32
	// indicate running -- even if not a service so we can use in loop below
	serviceStatus.dwCurrentState = SERVICE_RUNNING;
	serviceStatus.dwCheckPoint = 0;
	serviceStatus.dwWaitHint = 0;

	if (bAsWinService)
	{
		if (!SetServiceStatus(hServiceStatus, &serviceStatus))
		{
			sprintf(sReport,"ERROR calling SetServiceStatus: %d", GetLastError());
			report(sReport);
			return;
		}
	}

	// loop getting latest progress, then sending it as a udp msg
	while (serviceStatus.dwCurrentState == SERVICE_RUNNING)
#else
	// loop getting latest progress, then sending it as a udp msg
	while (TRUE)
#endif
	{
		// call fn to do a progress report
		doProgress();

		if (iCPURatingCheck-- < 1) iCPURatingCheck = 60; // reset to approx every 10 mins

		// sleep then do it again!
		SLEEP(cfgDelay);
	}

	close(sd);

#ifdef WIN32
	WSACleanup();
#endif
	if (bDebug) report("Shutdown.");
	return;
}

/**********************************************************************************/
// this fn gathers the progress information from various files, then sends a UDP
// datagram over to the progress system
int doProgress()
{
	char sFile[FILENAME_MAX]; // filename
	char sReport[FILENAME_MAX];
	char sDatLine[DAT_LINELEN + 1]; // 1 line from the file
	char sTemp[11]; // for converting string to num
	char sNodeName[32];
	char sProgress[256];
	char sCPURating[16];
	FILE *fp;
	int i;
	int iCompleted;
	char *spStatus;


	strcpy(sFile, cfgFADDir);  // start with dir
	strcat(sFile, "server.dat"); // append filename we want to open
	fp = fopen(sFile, "r"); // open readonly
	if (fp == NULL) // check for probs
	{
		//reportError(false, null, sFile + " not found");
		strcat(sProgress, sPACKET_TYPE_PROBLEM);
		strcat(sProgress, sPACKET_DELIM);
		sprintf(sReport, "WARNING:  error opening file %s.  Continuing...\n", sFile);
		report(sReport);
		return FALSE;
	}

	// header line
	memset(sDatLine, 0, sizeof(sDatLine));
	fgets(sDatLine, DAT_LINELEN, fp);
	memset(sTemp, 0, sizeof(sTemp));
	strncpy(sTemp, sDatLine + 20, 10); // 3rd value seems to be number of completed lines, i.e. one beyond is active target
	iCompleted = atoi(sTemp);

	// skip over completed targets
	for (i = 0; i < iCompleted; i++)
	{
		fgets(sDatLine, DAT_LINELEN, fp);
	}

	memset(sDatLine, 0, sizeof(sDatLine));
	do
	{
		sDatLine[84] = 0; // terminate the status field
		i = atoi(&sDatLine[82]); // convert status to int
		if (i == 5) // 5 is in progress it seems
		{
			// stuff delimiters into breakpoints within line
			sDatLine[32] = sDatLine[49] = sDatLine[65] = cPACKET_DELIM;
			sDatLine[81] = sDatLine[84] = sDatLine[89] = sDatLine[92] = cPACKET_DELIM;

			// grab node name out of the line
			memset(sNodeName, 0, sizeof(sNodeName));
			strncpy(sNodeName, &sDatLine[33], 16);
			trim(sNodeName);

			// get cpu rating sometimes
			memset(sCPURating, 0, sizeof(sCPURating));
			if (iCPURatingCheck < 1) getCPURating(sNodeName, sCPURating);  // approx every 10 mins get cpu rating

			// construct the string to send
			sprintf(sProgress, "%s%c%s%c%s%c%s%c%s%c%s%c",
				sPACKET_VERSION, cPACKET_DELIM,
				cfgThisName, cPACKET_DELIM,
				sNodeName, cPACKET_DELIM,
				sPACKET_TYPE_FAD, cPACKET_DELIM,
				sDatLine, cPACKET_DELIM,
				sCPURating, cPACKET_DELIM);

			// send the message -- might be a problem, might be normal
			sendto(sd, sProgress, (strlen(sProgress) + 1), 0, (const struct sockaddr *) &udpServer, sizeof(udpServer));
			if (bDebug)
			{
				sprintf(sReport, "sent to %s: %s", cfgServerName, sProgress);
				report(sReport);
			}
		}
		spStatus = fgets(sDatLine, DAT_LINELEN, fp);
	} while (spStatus != NULL); // loop until eof
	fclose(fp);
	return TRUE;
}

/**********************************************************************************/
/**********************************************************************************/
// Retrieve the CPU rating of this FAD node
void getCPURating(char *spNodeName, char *spCPURating)
{
	char sReport[FILENAME_MAX];
	char sFile[FILENAME_MAX];
	char sNodeName[FILENAME_MAX];
	char sLine[LINE_LEN];
	char sRating[LINE_LEN];
	char *spStatus;
	FILE *fp;

/*------using node.txt doesnt work when one directory is shared by multiple nodes
	sprintf(sFile, "%s%s", cfgFADDir, "node.txt");
	fp = fopen(sFile, "r"); // open readonly
	if (fp == NULL) // check for probs
	{
		//reportError(false, null, sFile + " not found");
		sprintf(sReport, "WARNING:  error opening file %s.  Continuing...\n", sFile);
		report(sReport);
		return;
	}
	// Read node name
	memset(sNodeName, 0, sizeof(sNodeName));
	fgets(sNodeName, FILENAME_MAX, fp);
	fclose(fp);
	sNodeName[strlen(sNodeName) - 1] = 0;
*/
	sprintf(sFile, "%s%s.tlg", cfgFADDir, spNodeName);
	fp = fopen(sFile, "r"); // open readonly
	if (fp == NULL) // check for probs
	{
		//reportError(false, null, sFile + " not found");
		sprintf(sReport, "WARNING:  error opening file %s.  Continuing...\n", sFile);
		report(sReport);
		return;
	}

	*sLine = 0;
	*sRating = 0;
	do
	{
		if (strstr(sLine, "CPU")) strcpy(sRating, sLine);

		// get next line from file
		spStatus = fgets(sLine, LINE_LEN, fp);
	} while (spStatus != NULL); // loop until eof
	fclose(fp);
	if (spStatus = strrchr(sRating, ':')) // locate colon
	{
		strcpy(spCPURating, spStatus + 1); // +1 skips over the colon
		trim(spCPURating); // remove leading and trailing white space
	}
}

/**********************************************************************************/
/**********************************************************************************/
// clean white space at start and end of string
char * trim(char *spLine)
{
	int i;
	while (isspace(*spLine) && *spLine != 0)
	{
		for (i = 0; i < strlen(spLine); i++) spLine[i] = spLine[i + 1];
	}
	i = strlen(spLine) - 1;
	while ((i > 0) && (isspace(spLine[i]))) spLine[i--] = 0;
	return spLine;
}

/**********************************************************************************/
// report this info to the console or log
int report(char * spLine)
{
	FILE* fp;
	time_t tm;
	char sTemp[32];
	char sLogfile[FILENAME_MAX];
	static int bDidShowHeader = { FALSE };

	// get and format current time
	tm = time(NULL);
	strftime(sTemp, sizeof(sTemp), sTIMEFORMAT, localtime(&tm));
	if (bUseLogfile)
	{
		strcpy(sLogfile, cfgLogDir);
		strcat(sLogfile, "/");
		strcat(sLogfile, sLOGFILENAME);
		fp = fopen(sLogfile, (bDidShowHeader ? "a":"w"));
		if (fp == NULL) return FALSE;

		// if we haven't already output the header, do so now
		if (! bDidShowHeader)
		{
			bDidShowHeader = TRUE;
			fprintf(fp, "%s\n", sPROGRAM_HEADER);
		}
		// finally -- output the line!
		fprintf(fp, "%s: %s\n", sTemp, spLine);
		fclose(fp);
	} else { // not using a log file
		// if we haven't already output the header, do so now
		if (! bDidShowHeader)
		{
			bDidShowHeader = TRUE;
			printf("%s\n", sPROGRAM_HEADER);
		}
		printf("%s: %s\n", sTemp, spLine);
	}
	return TRUE;
}


/**********************************************************************************/
// report this info to the console or log
void appendSeparator(char * spLine)
{
	// append trailing separator if necessary
	int iLen = strlen(spLine);
	if (iLen > 0)
	{
		if ((spLine[iLen - 1] != '\\') && (spLine[iLen - 1] != '/'))
			strcat(spLine, "/");
	}
}

/**********************************************************************************/
// read the fadprogressd.ini file for settings -- win svc mode only
void readIni(char * sIniFile)
{
	FILE *fp;
	char *spEquals;
	char *spStatus;
	char sLine[LINE_LEN];
	char sValue[LINE_LEN];

	fp = fopen(sIniFile, "r"); // open readonly
	if (fp != NULL) // only proceed if no probs; dont care if prob
	{
		*sLine = 0;
		do
		{
			spEquals = strchr(sLine, '=');
			if (spEquals != NULL)
			{
				strcpy(sValue, spEquals); // from equals sign beyond copied to sValue
				*spEquals = 0; // terminate sLine in front of the equals sign
				if (strlen(sValue) != 0)
				{
					sValue[0] = ' '; // replace the equals sign with a space
					trim(sValue); // trim out the space

					if (0 == strcmp(sLine, "sendto")) {
						strcpy(cfgServerName, sValue);
					} else if (0 == strcmp(sLine, "dir")) {
						strcpy(cfgFADDir, sValue);
						appendSeparator(cfgFADDir);
					} else if (0 == strcmp(sLine, "logdir")) {
						strcpy(cfgLogDir, sValue);
						appendSeparator(cfgLogDir);
					} else if (0 == strcmp(sLine, "name")) {
						strcpy(cfgThisName, sValue);
					} else if (0 == strcmp(sLine, "port")) {
						cfgServerPort = atoi(sValue);
					} else if (0 == strcmp(sLine, "timer")) {
						cfgDelay = atoi(sValue);
					} else if (0 == strcmp(sLine, "debug")) {
						bDebug = TRUE;
					}
				}
			}

			// get next line from file
			spStatus = fgets(sLine, LINE_LEN, fp);
		} while (spStatus != NULL); // loop until eof
		fclose(fp);
	}
	return;
}


/**********************************************************************************/
#ifdef WIN32
/**********************************************************************************/
// this handler communicates with the SCM to control activity
void WINAPI serviceControlHandler(DWORD iOpcode)
{
	DWORD status;

	switch(iOpcode)
	{
		case SERVICE_CONTROL_STOP:
			// by setting the state to stopped, it will stop the serviceMain thread too
			serviceStatus.dwCurrentState = SERVICE_STOPPED;
			serviceStatus.dwWin32ExitCode = 0;
			serviceStatus.dwCheckPoint = 0;
			serviceStatus.dwWaitHint = 0;
			if (bDebug) report("Leaving service.");
			break;

		case SERVICE_CONTROL_INTERROGATE:
			// send current status
			break;
	}

	if (!SetServiceStatus(hServiceStatus, &serviceStatus))
	{
		char sReport[256];
		sprintf(sReport, "ERROR from SetServiceStatus: %d", GetLastError());
		report(sReport);
	}
	return;
}

/**********************************************************************************/
// stop the service
void stopService()
{
	serviceStatus.dwCurrentState = SERVICE_STOPPED;
	serviceStatus.dwCheckPoint = 0;
	serviceStatus.dwWaitHint = 0;
	serviceStatus.dwWin32ExitCode = 0;
	serviceStatus.dwServiceSpecificExitCode = 0;

	SetServiceStatus(hServiceStatus, &serviceStatus);
	return;
}


/**********************************************************************************/
// create the windows service
void createService()
{
	TCHAR szApplication[MAX_PATH];
	char sReport[256];
	SC_HANDLE hService;
	SC_HANDLE hServiceManager;

	// tell user
	sprintf(sReport, "Creating service \"%s\" (%s)", sSERVICE_DISPLAY, sSERVICE_NAME);
	report(sReport);
	// get app name
	if (!GetModuleFileName(NULL, szApplication, MAX_PATH))
	{
		sprintf(sReport, "ERROR from GetModuleFileName: %d", GetLastError());
		report(sReport);
		return;
	}
	// open mgr
	hServiceManager = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);
	if (hServiceManager == NULL)
	{
		sprintf(sReport, "ERROR from OpenSCManager: %d", GetLastError());
		report(sReport);
		return;
	}
	// create svc
	hService = CreateService(hServiceManager,
		TEXT(sSERVICE_NAME),
		TEXT(sSERVICE_DISPLAY),
		SERVICE_ALL_ACCESS,
		SERVICE_WIN32_OWN_PROCESS,
		SERVICE_AUTO_START,
		SERVICE_ERROR_NORMAL,
		szApplication,
		NULL, NULL, NULL, NULL, NULL);
	if (hService == NULL)
	{
		sprintf(sReport, "ERROR from CreateService: %d", GetLastError());
		report(sReport);
		CloseServiceHandle(hServiceManager);
		return;
	}

	// Created.
	report("Service created, starting...");

	// Try to start it
	if (! StartService(hService, 0, NULL))
	{
		sprintf(sReport, "ERROR from StartService: %d", GetLastError());
		report(sReport);
		CloseServiceHandle(hService);
		CloseServiceHandle(hServiceManager);
		return;
	}
	report("Service told to start.");

	// cleanup
	CloseServiceHandle(hService);
	CloseServiceHandle(hServiceManager);
}

/**********************************************************************************/
// delete the windows service
void deleteService()
{
	TCHAR szApplication[MAX_PATH];
	char sReport[256];
	SC_HANDLE hService;
	SC_HANDLE hServiceManager;
	SERVICE_STATUS lpSvcStatus;

	// tell user
	sprintf(sReport, "Stopping service \"%s\" (%s)", sSERVICE_DISPLAY, sSERVICE_NAME);
	report(sReport);
	// open mgr
	hServiceManager = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);
	if (hServiceManager == NULL)
	{
		sprintf(sReport, "ERROR from OpenSCManager: %d", GetLastError());
		report(sReport);
		return;
	}
	// open svc
	hService = OpenService(hServiceManager, sSERVICE_NAME, DELETE | SERVICE_STOP);
	if (hService == NULL)
	{
		sprintf(sReport, "ERROR from OpenService: %d", GetLastError());
		report(sReport);
		CloseServiceHandle(hServiceManager);
		return;
	}
	// stop svc
	if (! ControlService(hService, SERVICE_CONTROL_STOP, &lpSvcStatus))
	{
		sprintf(sReport, "ERROR from ControlService to stop: %d", GetLastError());
		report(sReport);
		report("Continuing.");
	}
	// delete svc
	report("Deleting service...");
	if (! DeleteService(hService))
	{
		sprintf(sReport, "ERROR from DeleteService: %d", GetLastError());
		report(sReport);
		CloseServiceHandle(hService);
		CloseServiceHandle(hServiceManager);
		return;
	}

	// Deleted.  now closeup.
	report("Service deleted.");
	CloseServiceHandle(hService);
	CloseServiceHandle(hServiceManager);
}

/**********************************************************************************/
#endif // WIN32
/**********************************************************************************/
